import pandas as pd
import plorts
import numpy as np
import matplotlib.pyplot as plt
import re
import sys
import os
import gzip
from collections import defaultdict
import tcp_probe.parser as parser
plt.style.use(['plorts','plorts-print'])

def plot_trace(output_dir, trace_path, normalize_time):
    trace_dir = os.path.dirname(trace_path)
    trace_file = os.path.basename(trace_path)

    lines = parser.open_trace(trace_path)

    rows = [parser.parse_tcp_probe_line(line) for line in lines]
    rows = [r for r in rows if r is not None]
    df = pd.DataFrame(rows)

    xmin,xmax=df.timestamp.min(), df.timestamp.max()

    if normalize_time:
        df['timestamp'] = df['timestamp'] - xmin
        xlabel = "Timestamp (sec)"
    else:
        xlabel = "Timestamp"
        
    num_plots = 3
    
    plt.figure(figsize=(18,15))
    plt.subplot(num_plots,1,1)
    plt.title("snd_cwnd")
    plorts.scatter(df, x="timestamp", y="snd_cwnd", hue=["sport", "dport"])
    plt.xlabel(xlabel)
    plt.legend(loc='best')

    plt.subplot(num_plots,1,2)
    plt.title("srtt")
    plorts.scatter(df, x="timestamp", y="srtt", hue=["sport", "dport"])
    plt.axis(xmin=xmin, xmax=xmax)
    plt.xlabel(xlabel)
    plt.legend(loc='best')

    plt.subplot(num_plots,1,3)
    rows = [parser.parse_tcp_retransmit_skb_line(line) for line in lines]
    rows = [r for r in rows if r is not None]
    df = pd.DataFrame(rows)
    
    if len(df) > 0:
        if normalize_time:
            df['timestamp'] = df['timestamp'] - xmin

        flow_ids = {}
        for sport in df.sport.unique():
            flow_ids[sport] = len(flow_ids)
        df["flow_id"] = [flow_ids[sport] for sport in df.sport]

        plt.title("retransmitted skbs")
        plorts.scatter(df, x="timestamp", y="flow_id", hue="sport")
        plt.axis(xmin=xmin, xmax=xmax)
        plt.xlabel(xlabel)

    if output_dir is None:
        output_path = '%s.png'%(trace_path)
    else:
        output_path = os.path.join(trace_dir, output_dir, '%s.png'%(trace_file))

    plt.savefig(output_path, dpi=300, bbox_inches="tight")



#!/usr/bin/env python

import sys
import os
import subprocess
import click

def enable_event(event):
    f = open("/sys/kernel/debug/tracing/events/%s/enable"%event, mode='w')
    f.write('1')
    f.close()

def disable_event(event):
    f = open("/sys/kernel/debug/tracing/events/%s/enable"%event, mode='w')
    f.write('0')
    f.close()
    
def set_trace_buffer_size(buffer_size_kb):
    subprocess.call("echo %d > /sys/kernel/debug/tracing/buffer_size_kb"%buffer_size_kb,
                    shell=True)

def set_trace_filter(event, filter_string):
    f = open("/sys/kernel/debug/tracing/events/%s/filter"%event, mode='w')
    f.write(filter_string)
    f.close()

def clear_trace_buffer():
    subprocess.call('echo "" > /sys/kernel/debug/tracing/trace', shell=True)

def validate_filter(ctx, param, filter_string):
    # TODO: actually validate the filter string
    return filter_string

@click.group()
def cli():
    pass

@cli.command()
@click.option('--output-dir', default=None, type=click.Path(exists=True))
@click.option('--normalize-time/--raw-time', default=True, help="Show time from beginning of trace/wall-clock time")
@click.argument('trace_path', metavar='trace', type=click.Path(exists=True))
def plot(output_dir, normalize_time, trace_path):
    """Generate plots from a TCP trace"""
    import tcp_probe.plot
    tcp_probe.plot.plot_trace(output_dir, trace_path, normalize_time)
    
@cli.command()
@click.option('--clear/--no-clear', default=True, help="Whether to clear the kernel trace buffer before starting the trace")
@click.option('-b', '--buffer', 'buffer_size', default=500000, help="Size of trace buffer in kb")
@click.argument('filter_string', metavar='filter', default='', callback=validate_filter)
def trace(clear, buffer_size, filter_string):
    """Takes a tcp-related trace from the Linux kernel.

    Takes an optional filter for the kernel trace. Please see the kernel documentation for syntax: https://www.kernel.org/doc/html/latest/trace/events.html#event-filtering
    """
    try:
        enable_event("tcp")
        # enable_event("tcp/tcp_probe")
        # enable_event("tcp/tcp_retransmit_skb")
        # enable_event("tcp/tcp_retransmit_synack")
    except PermissionError as e:
        print("error: insufficient permissions to enable event tracing", file=sys.stderr)
        sys.exit(1)

    set_trace_buffer_size(buffer_size)
    set_trace_filter("tcp", filter_string)

    if clear:
        clear_trace_buffer()

    try:
        subprocess.call('cat /sys/kernel/debug/tracing/trace_pipe', shell=True)
    finally:
        disable_event("tcp")
        set_trace_filter("tcp", "")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tcp_probe", # Replace with your own username
    version="0.0.1",
    author="Bruce Spang",
    author_email="bruce@brucespang.com",
    description="Tools for tracing TCP in Linux",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/brucespang/tcp_probe",
    packages=setuptools.find_packages(),
    install_requires=[
        'Click',
        # technically yes, but creates annoying dependency issues on servers...
        # 'pandas'
        # 'plorts'
        # 'matplotlib'
        # 'numpy'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Operating System Kernels :: Linux",
        "Topic :: System :: Networking :: Monitoring"
    ],
    python_requires='>=3',
    entry_points={
        'console_scripts': [
            "tcp_probe = tcp_probe.cli:cli",
        ]
    }
)



import re
import sys
import os
import gzip

def open_trace(f):
    path = os.path.dirname(f)
    filename = os.path.basename(f)
    if '.gz' in filename:
        lines = list(gzip.open(f))
    else:
        lines = list(open(f))
    return lines

def parse_tcp_probe_line(line):
    if "tcp_probe" not in line:
        return None
    
    match = re.match(".*\s(.*): tcp_probe: (.*)", line.strip())
    if match is None:
        return None
    
    timestamp = float(match.group(1))
    data = {pair.split('=')[0]:pair.split('=')[1] for pair in match.group(2).split(' ')}
    integer_fields = ['rcv_wnd', 'snd_cwnd', 'snd_wnd', 'srtt', 'ssthresh', 'data_len']
    for f in integer_fields:
        try:
            data[f] = int(data[f])
        except:
            print(f)
            
    hex_fields = ['snd_nxt', 'snd_una', 'mark']
    for f in hex_fields:
        data[f] = int(data[f], 16)
        data['timestamp'] = timestamp

    data['sport'] = int(data['src'].split(':')[-1])
    data['dport'] = int(data['dest'].split(':')[-1])
    
    return data

def parse_tcp_retransmit_skb_line(line):
    if "tcp_retransmit_skb" not in line:
        return None
    
    match = re.match(".*\s(.*): tcp_retransmit_skb: (.*)", line.strip())
    if match is None:
        return None
    
    timestamp = float(match.group(1))
    data = {pair.split('=')[0]:pair.split('=')[1] for pair in match.group(2).split(' ')}
    integer_fields = ['sport', 'dport']
    for f in integer_fields:
        try:
            data[f] = int(data[f])
        except:
            print(f)
    data['timestamp'] = timestamp
    return data
